package java.util;

import java.util.function.Consumer;
import sun.misc.SharedSecrets;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;

public class PriorityQueue<E> extends AbstractQueue<E> implements Serializable
{
    private static final long serialVersionUID = -7720805057305804111L;
    private static final int DEFAULT_INITIAL_CAPACITY = 11;
    transient Object[] queue;
    private int size;
    private final Comparator<? super E> comparator;
    transient int modCount;
    private static final int MAX_ARRAY_SIZE = 2147483639;
    
    public PriorityQueue() {
        this(11, null);
    }
    
    public PriorityQueue(int n) {
        this(n, null);
    }
    
    public PriorityQueue(Comparator<? super E> comparator) {
        this(11, comparator);
    }
    
    public PriorityQueue(int n, Comparator<? super E> comparator) {
        this.size = 0;
        this.modCount = 0;
        if (n < 1) {
            throw new IllegalArgumentException();
        }
        this.queue = new Object[n];
        this.comparator = comparator;
        return;
    }
    
    public PriorityQueue(Collection<? extends E> collection) {
        this.size = 0;
        this.modCount = 0;
        if ((collection instanceof SortedSet)) {
            SortedSet<Object> set = (SortedSet)collection;
            this.comparator = set.comparator();
            this.initElementsFromCollection(set);
        }
        else if ((collection instanceof PriorityQueue)) {
            PriorityQueue<Object> priorityQueue = (PriorityQueue)collection;
            this.comparator = priorityQueue.comparator();
            this.initFromPriorityQueue(priorityQueue);
        }
        else {
            this.comparator = null;
            this.initFromCollection(collection);
        }
    }
    
    public PriorityQueue(PriorityQueue<? extends E> priorityQueue) {
        this.size = 0;
        this.modCount = 0;
        this.comparator = priorityQueue.comparator();
        this.initFromPriorityQueue(priorityQueue);
    }
    
    public PriorityQueue(SortedSet<? extends E> set) {
        this.size = 0;
        this.modCount = 0;
        this.comparator = set.comparator();
        this.initElementsFromCollection(set);
    }
    
    private void initFromPriorityQueue(PriorityQueue<? extends E> priorityQueue) {
        if (priorityQueue.getClass() == PriorityQueue.class) {
            this.queue = priorityQueue.toArray();
            this.size = priorityQueue.size();
        }
        else {
            this.initFromCollection(priorityQueue);
        }
    }
    
    private void initElementsFromCollection(Collection<? extends E> collection) {
        Object[] queue = collection.toArray();
        if (collection.getClass() != ArrayList.class) {
            queue = Arrays.copyOf(queue, queue.length, Object[].class);
        }
        int length = queue.length;
        if (length == 1 || this.comparator != null) {
            int i = 0;
            while (i < length) {
                if (queue[i] == null) {
                    throw new NullPointerException();
                }
                ++i;
            }
        }
        this.queue = queue;
        this.size = queue.length;
    }
    
    private void initFromCollection(Collection<? extends E> collection) {
        this.initElementsFromCollection(collection);
        this.heapify();
    }
    
    private void grow(int n) {
        int length = this.queue.length;
        int hugeCapacity = length + ((length < 64) ? (length + 2) : (length >> 1));
        if (hugeCapacity - 2147483639 > 0) {
            hugeCapacity = hugeCapacity(n);
        }
        this.queue = Arrays.copyOf(this.queue, hugeCapacity);
    }
    
    private static int hugeCapacity(int n) {
        if (n < 0) {
            throw new OutOfMemoryError();
        }
        return (n > 2147483639) ? Integer.MAX_VALUE : 2147483639;
    }
    
    @Override
    public boolean add(E e) {
        return this.offer(e);
    }
    
    @Override
    public boolean offer(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        ++this.modCount;
        int size = this.size;
        if (size >= this.queue.length) {
            this.grow(size + 1);
        }
        this.size = size + 1;
        if (size == 0) {
            this.queue[0] = e;
        }
        else {
            this.siftUp(size, e);
        }
        return true;
    }
    
    @Override
    public E peek() {
        return (this.size == 0) ? null : this.queue[0];
    }
    
    private int indexOf(Object o) {
        if (o != null) {
            int i = 0;
            while (i < this.size) {
                if (o.equals(this.queue[i])) {
                    return i;
                }
                ++i;
            }
        }
        return -1;
    }
    
    @Override
    public boolean remove(Object o) {
        int index = this.indexOf(o);
        if (index == -1) {
            return false;
        }
        this.removeAt(index);
        return true;
    }
    
    boolean removeEq(Object o) {
        int i = 0;
        while (i < this.size) {
            if (o == this.queue[i]) {
                this.removeAt(i);
                return true;
            }
            ++i;
        }
        return false;
    }
    
    @Override
    public boolean contains(Object o) {
        return this.indexOf(o) != -1;
    }
    
    @Override
    public Object[] toArray() {
        return Arrays.copyOf(this.queue, this.size);
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        int size = this.size;
        if (array.length < size) {
            return (T[])Arrays.copyOf(this.queue, size, array.getClass());
        }
        System.arraycopy(this.queue, 0, array, 0, size);
        if (array.length > size) {
            array[size] = null;
            return array;
        }
        return array;
    }
    
    @Override
    public Iterator<E> iterator() {
        return new PriorityQueue.Itr();
    }
    
    @Override
    public int size() {
        return this.size;
    }
    
    @Override
    public void clear() {
        ++this.modCount;
        for (int i = 0; i < this.size; ++i) {
            this.queue[i] = null;
        }
        this.size = 0;
    }
    
    @Override
    public E poll() {
        if (this.size == 0) {
            return null;
        }
        int size = this.size - 1;
        this.size = size;
        int n = size;
        ++this.modCount;
        Object o = this.queue[0];
        Object o2 = this.queue[n];
        this.queue[n] = null;
        if (n != 0) {
            this.siftDown(0, o2);
            return o;
        }
        return o;
    }
    
    private E removeAt(int n) {
        ++this.modCount;
        int size = this.size - 1;
        this.size = size;
        int n2 = size;
        if (n2 == n) {
            this.queue[n] = null;
        }
        else {
            Object o = this.queue[n2];
            this.queue[n2] = null;
            this.siftDown(n, o);
            if (this.queue[n] == o) {
                this.siftUp(n, o);
                if (this.queue[n] != o) {
                    return o;
                }
            }
        }
        return null;
    }
    
    private void siftUp(int n, E e) {
        if (this.comparator != null) {
            this.siftUpUsingComparator(n, e);
        }
        else {
            this.siftUpComparable(n, e);
        }
    }
    
    private void siftUpComparable(int i, E e) {
        Comparable<Object> comparable = (Comparable)e;
        while (i > 0) {
            int n = i - 1 >>> 1;
            Object o = this.queue[n];
            if (comparable.compareTo(o) >= 0) {
                break;
            }
            this.queue[i] = o;
            i = n;
        }
        this.queue[i] = comparable;
    }
    
    private void siftUpUsingComparator(int i, E e) {
        while (i > 0) {
            int n = i - 1 >>> 1;
            Object o = this.queue[n];
            if (this.comparator.compare(e, o) >= 0) {
                break;
            }
            this.queue[i] = o;
            i = n;
        }
        this.queue[i] = e;
    }
    
    private void siftDown(int n, E e) {
        if (this.comparator != null) {
            this.siftDownUsingComparator(n, e);
        }
        else {
            this.siftDownComparable(n, e);
        }
    }
    
    private void siftDownComparable(int i, E e) {
        Comparable<Object> comparable = (Comparable)e;
        while (i < this.size >>> 1) {
            int n = (i << 1) + 1;
            Object o = this.queue[n];
            int n2 = n + 1;
            if (n2 < this.size && ((Comparable)o).compareTo(this.queue[n2]) > 0) {
                o = this.queue[n = n2];
            }
            if (comparable.compareTo(o) <= 0) {
                break;
            }
            this.queue[i] = o;
            i = n;
        }
        this.queue[i] = comparable;
    }
    
    private void siftDownUsingComparator(int i, E e) {
        while (i < this.size >>> 1) {
            int n = (i << 1) + 1;
            Object o = this.queue[n];
            int n2 = n + 1;
            if (n2 < this.size && this.comparator.compare(o, this.queue[n2]) > 0) {
                o = this.queue[n = n2];
            }
            if (this.comparator.compare(e, o) <= 0) {
                break;
            }
            this.queue[i] = o;
            i = n;
        }
        this.queue[i] = e;
    }
    
    private void heapify() {
        for (int i = (this.size >>> 1) - 1; i >= 0; --i) {
            this.siftDown(i, this.queue[i]);
        }
    }
    
    public Comparator<? super E> comparator() {
        return this.comparator;
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        objectOutputStream.writeInt(Math.max(2, this.size + 1));
        for (int i = 0; i < this.size; ++i) {
            objectOutputStream.writeObject(this.queue[i]);
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        objectInputStream.readInt();
        SharedSecrets.getJavaOISAccess().checkArray(objectInputStream, Object[].class, this.size);
        this.queue = new Object[this.size];
        for (int i = 0; i < this.size; ++i) {
            this.queue[i] = objectInputStream.readObject();
        }
        this.heapify();
    }
    
    @Override
    public final Spliterator<E> spliterator() {
        return new PriorityQueue.PriorityQueueSpliterator(this, 0, -1, 0);
    }
    
    static Object access$200(PriorityQueue priorityQueue, int n) {
        return priorityQueue.removeAt(n);
    }
    
    private final class Itr implements Iterator<E>
    {
        private int cursor;
        private int lastRet;
        private ArrayDeque<E> forgetMeNot;
        private E lastRetElt;
        private int expectedModCount;
        
        private Itr() {
            this.cursor = 0;
            this.lastRet = -1;
            this.forgetMeNot = null;
            this.lastRetElt = null;
            this.expectedModCount = PriorityQueue.this.modCount;
        }
        
        @Override
        public boolean hasNext() {
            return this.cursor < PriorityQueue.this.size || (this.forgetMeNot != null && !this.forgetMeNot.isEmpty());
        }
        
        @Override
        public E next() {
            if (this.expectedModCount != PriorityQueue.this.modCount) {
                throw new ConcurrentModificationException();
            }
            if (this.cursor < PriorityQueue.this.size) {
                Object[] queue = PriorityQueue.this.queue;
                int lastRet = this.cursor++;
                this.lastRet = lastRet;
                return queue[lastRet];
            }
            if (this.forgetMeNot != null) {
                this.lastRet = -1;
                this.lastRetElt = this.forgetMeNot.poll();
                if (this.lastRetElt != null) {
                    return this.lastRetElt;
                }
            }
            throw new NoSuchElementException();
        }
        
        @Override
        public void remove() {
            if (this.expectedModCount != PriorityQueue.this.modCount) {
                throw new ConcurrentModificationException();
            }
            if (this.lastRet != -1) {
                Object access$200 = PriorityQueue.access$200(PriorityQueue.this, this.lastRet);
                this.lastRet = -1;
                if (access$200 == null) {
                    --this.cursor;
                }
                else {
                    if (this.forgetMeNot == null) {
                        this.forgetMeNot = new ArrayDeque();
                    }
                    this.forgetMeNot.add(access$200);
                }
            }
            else if (this.lastRetElt != null) {
                PriorityQueue.this.removeEq(this.lastRetElt);
                this.lastRetElt = null;
            }
            else {
                throw new IllegalStateException();
            }
            this.expectedModCount = PriorityQueue.this.modCount;
            return;
        }
    }
    
    static final class PriorityQueueSpliterator<E> implements Spliterator<E>
    {
        private final PriorityQueue<E> pq;
        private int index;
        private int fence;
        private int expectedModCount;
        
        PriorityQueueSpliterator(PriorityQueue<E> pq, int index, int fence, int expectedModCount) {
            this.pq = pq;
            this.index = index;
            this.fence = fence;
            this.expectedModCount = expectedModCount;
        }
        
        private int getFence() {
            int fence;
            if ((fence = this.fence) < 0) {
                this.expectedModCount = this.pq.modCount;
                int access$100 = this.pq.size;
                this.fence = access$100;
                fence = access$100;
            }
            return fence;
        }
        
        @Override
        public PriorityQueue.PriorityQueueSpliterator<E> trySplit() {
            int fence = this.getFence();
            int index = this.index;
            int index2 = index + fence >>> 1;
            return (index >= index2) ? null : new PriorityQueue.PriorityQueueSpliterator<E>(this.pq, index, this.index = index2, this.expectedModCount);
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            PriorityQueue<Object> pq;
            Object[] queue;
            if ((pq = this.pq) != null && (queue = pq.queue) != null) {
                int index;
                int n;
                if ((index = this.fence) < 0) {
                    n = pq.modCount;
                    index = pq.size;
                }
                else {
                    n = this.expectedModCount;
                }
                int i;
                if ((i = this.index) >= 0 && (this.index = index) <= queue.length) {
                    while (i < index) {
                        Object o;
                        if ((o = queue[i]) == null) {
                            throw new ConcurrentModificationException();
                        }
                        consumer.accept(o);
                        ++i;
                    }
                    if (pq.modCount == n) {
                        return;
                    }
                }
            }
            throw new ConcurrentModificationException();
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            int fence = this.getFence();
            int index = this.index;
            if (index >= 0 && index < fence) {
                this.index = index + 1;
                Object o = this.pq.queue[index];
                if (o == null) {
                    throw new ConcurrentModificationException();
                }
                consumer.accept(o);
                if (this.pq.modCount != this.expectedModCount) {
                    throw new ConcurrentModificationException();
                }
                return true;
            }
            else {
                return false;
            }
        }
        
        @Override
        public long estimateSize() {
            return (long)(this.getFence() - this.index);
        }
        
        @Override
        public int characteristics() {
            return 16704;
        }
    }
}
